@@ -1,8 +1,13 @@
+from __future__ import annotations
+
 import asyncio
 import logging
 from contextlib import suppress
+from ipaddress import IPv4Address
+from typing import cast
 
 from dns import exception, message, rdatatype
+from dns.message import Message
 
 from .network import SystemNetworkData
 
@@ -18,12 +23,12 @@ DNS_PORT = 53
 _LOGGER = logging.getLogger(__name__)
 
 
-def short_hostname(hostname):
+def short_hostname(hostname: str) -> str:
     """The first part of the hostname."""
     return hostname.split(".")[0]
 
 
-def dns_message_short_hostname(dns_message):
+def dns_message_short_hostname(dns_message: message.Message | None) -> str | None:
     """Get the short hostname from a dns message."""
     if not isinstance(dns_message, message.Message) or not dns_message.answer:
         return None
@@ -38,22 +43,22 @@ def dns_message_short_hostname(dns_message):
 class PTRResolver:
     """Implement DNS PTR resolver."""
 
-    def __init__(self, destination):
+    def __init__(self, destination: tuple[str, int]) -> None:
         """Init protocol for a destination."""
-        self.destination = destination
-        self.responses = {}
-        self.send_id = None
+        self.destination: tuple[str, int] = destination
+        self.responses: dict[int, Message] = {}
+        self.send_id: int | None = None
         self.responded = asyncio.Event()
-        self.error = None
+        self.error: Exception | None = None
 
-    def connection_lost(self, transport):
+    def connection_lost(self, ex: Exception | None) -> None:
         """Connection lost."""
 
-    def connection_made(self, transport):
+    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
         """Connection made."""
         self.transport = transport
 
-    def datagram_received(self, data, addr):
+    def datagram_received(self, data: bytes, addr: tuple[str, int]) -> None:
         """Response recieved."""
         if addr != self.destination:
             return
@@ -65,12 +70,12 @@ class PTRResolver:
         if msg.id == self.send_id:
             self.responded.set()
 
-    def error_received(self, exc):
+    def error_received(self, exc: Exception) -> None:
         """Error received."""
         self.error = exc
         self.responded.set()
 
-    async def send_query(self, query):
+    async def send_query(self, query: Message) -> None:
         """Send a query and wait for a response."""
         self.responded.clear()
         self.send_id = query.id
@@ -80,14 +85,18 @@ class PTRResolver:
             raise self.error
 
 
-async def async_query_for_ptrs(nameserver, ips_to_lookup):
+async def async_query_for_ptrs(
+    nameserver: str, ips_to_lookup: list[IPv4Address]
+) -> list[Message | None]:
     """Fetch PTR records for a list of ips."""
     destination = (nameserver, DNS_PORT)
     loop = asyncio.get_running_loop()
 
-    transport, protocol = await loop.create_datagram_endpoint(
-        lambda: PTRResolver(destination), remote_addr=destination
+    transport_protocol = await loop.create_datagram_endpoint(
+        lambda: PTRResolver(destination), remote_addr=destination  # type: ignore
     )
+    transport = cast(asyncio.DatagramTransport, transport_protocol[0])
+    protocol = cast(PTRResolver, transport_protocol[1])
     try:
         return await async_query_for_ptr_with_proto(protocol, ips_to_lookup)
     finally:
@@ -99,7 +108,9 @@ def async_generate_ptr_query(ip):
     return message.make_query(ip.reverse_pointer, rdatatype.PTR)
 
 
-async def async_query_for_ptr_with_proto(protocol, ips_to_lookup):
+async def async_query_for_ptr_with_proto(
+    protocol: PTRResolver, ips_to_lookup: list[IPv4Address]
+) -> list[Message | None]:
     """Send and receiver the PTR queries."""
     time_outs = 0
     query_for_ip = {}
@@ -117,26 +128,34 @@ async def async_query_for_ptr_with_proto(protocol, ips_to_lookup):
         except OSError:
             break
 
-    return [protocol.responses.get(query_for_ip.get(ip)) for ip in ips_to_lookup]
+    return [protocol.responses.get(query_for_ip[ip]) for ip in ips_to_lookup]
 
 
 class DiscoverHosts:
     """Discover hosts on the network by ARP and PTR lookup."""
 
-    def __init__(self):
+    def __init__(self) -> None:
         """Init the discovery hosts."""
-        self.ip_route = None
-        with suppress(Exception):
-            from pyroute2 import IPRoute  # pylint: disable=import-outside-toplevel
+        self._ip_route = None
 
-            self.ip_route = IPRoute()
+    def _get_sys_network_data(self) -> SystemNetworkData:
+        if not self._ip_route:
+            with suppress(Exception):
+                from pyroute2.iproute import (
+                    IPRoute,
+                )  # type: ignore # pylint: disable=import-outside-toplevel
 
-    async def async_discover(self):
+                self._ip_route = IPRoute()
+        sys_network_data = SystemNetworkData(self._ip_route)
+        sys_network_data.setup()
+        return sys_network_data
+
+    async def async_discover(self) -> list[dict[str, str]]:
         """Discover hosts on the network by ARP and PTR lookup."""
-        sys_network_data = SystemNetworkData(self.ip_route)
         loop = asyncio.get_running_loop()
-        await loop.run_in_executor(None, sys_network_data.setup)
+        sys_network_data = await loop.run_in_executor(None, self._get_sys_network_data)
         network = sys_network_data.network
+        assert network is not None
         if network.num_addresses > MAX_ADDRESSES:
             _LOGGER.debug(
                 "The network %s exceeds the maximum number of addresses, %s; No scanning performed",
@@ -156,21 +175,27 @@ class DiscoverHosts:
             if ip in neighbours
         ]
 
-    async def _async_get_nameservers(self, sys_network_data):
+    async def _async_get_nameservers(
+        self, sys_network_data: SystemNetworkData
+    ) -> list[str]:
         """Get nameservers to query."""
         all_nameservers = list(sys_network_data.nameservers)
         router_ip = sys_network_data.router_ip
+        assert router_ip is not None
         if router_ip not in all_nameservers:
             neighbours = await sys_network_data.async_get_neighbors([router_ip])
             if router_ip in neighbours:
                 all_nameservers.insert(0, router_ip)
         return all_nameservers
 
-    async def async_get_hostnames(self, sys_network_data):
+    async def async_get_hostnames(
+        self, sys_network_data: SystemNetworkData
+    ) -> dict[str, str]:
         """Lookup PTR records for all addresses in the network."""
         all_nameservers = await self._async_get_nameservers(sys_network_data)
+        assert sys_network_data.network is not None
         ips = list(sys_network_data.network.hosts())
-        hostnames = {}
+        hostnames: dict[str, str] = {}
         for nameserver in all_nameservers:
             ips_to_lookup = [ip for ip in ips if str(ip) not in hostnames]
             results = await async_query_for_ptrs(nameserver, ips_to_lookup)
